#!/usr/bin/env python
import os
import shutil
from .utils import (
    get_python_cmd,
    remove_compiled_prefix,
    generate_hps,
    show_strategies,
    remove_first_line_in_file,
    DEBUG_LEVEL_INFO
)
from .parser import get_parser
from .launch import launch

def main():
    parser, args = get_parser()

    if args.command is None:
        parser.print_help()
        exit(0)

    import torch

    if args.command in ["launch", "debug"]:
        if not args.suppress_queue:
            if not os.path.exists(args.queue_file):
                open(args.queue_file, "w").close()  # creates an emtpy file
            
            launch(args)
            if os.path.exists(args.queue_file):
                cmds = open(args.queue_file).read().splitlines()
                if len(cmds) > 0:
                    remove_first_line_in_file(args.queue_file)
                    os.system(cmds[0])
        else:
            launch(args)
    elif args.command == "get":
        assert args.out is not None, "You must specify an output directory ('--out')."
        assert hasattr(torch, args.dtype), f"'{args.dtype}' not supported in PyTorch."
        CHKPT_BASE_DIRECTORY = f"{args.checkpoint}/checkpoint"
        checkpoint_dir = CHKPT_BASE_DIRECTORY if os.path.exists(CHKPT_BASE_DIRECTORY) else args.get
        files = os.listdir(checkpoint_dir)

        python_cmd = get_python_cmd()
        os.makedirs(args.out, exist_ok=True)
        if "status.json" in os.listdir(args.checkpoint):
            shutil.copy(f"{args.checkpoint}/status.json", args.out)
        
        state_dict_file = f"{args.out}/pytorch_model.pt"

        if "zero_to_fp32.py" in files: # check for DeepSpeed
            print("Converting Zero to float32 parameters...")
            exit_code = os.system(f"{python_cmd} {checkpoint_dir}/zero_to_fp32.py {checkpoint_dir} {state_dict_file}")
            if exit_code != 0:
                raise RuntimeError("Something went wrong when converting Zero to float32.")
        elif "pytorch_model_fsdp_0" in files: # check for FSDP
            # using Accelerate's approach for now, and only checking for one node
            exit_code = os.system(f"accelerate merge-weights {checkpoint_dir}/pytorch_model_fsdp_0 {args.out}")
            if exit_code != 0:
                raise RuntimeError("Something went wrong when merging weights from FSDP.")
            
            shutil.copy(f"{checkpoint_dir}/pytorch_model.bin", f"{args.out}/pytorch_model.pt")
        else: # check for DDP
            shutil.copy(f"{checkpoint_dir}/pytorch_model.bin", f"{args.out}/pytorch_model.pt")
            
        state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)
        state_dict = remove_compiled_prefix(state_dict)

        _dtype_str = f" and converting to dtype {args.dtype}" if args.dtype is not None else ""
        print(f"Setting 'requires_grad' to False{_dtype_str}...")
        for key in state_dict.keys():
            state_dict[key].requires_grad = False
            if args.dtype is not None:
                state_dict[key] = state_dict[key].to(getattr(torch, args.dtype))

        torch.save(state_dict, state_dict_file)
        print(f"Model directory saved to '{args.out}'.")
    elif args.command == "strats":
        if args.ddp:
            show_strategies(filter="ddp")
        elif args.fsdp:
            show_strategies(filter="fsdp")
        elif args.deepspeed:
            show_strategies(filter="deepspeed")
        else:
            show_strategies()
    elif args.command == "example":
        generate_hps()
        print("'hps_example.yaml' generated.")
    elif args.command == "debug-levels":
        if args.level is None:
            for level, info in DEBUG_LEVEL_INFO.items():
                print(f"  Level {level}: {info}")
        else:
            if args.level in DEBUG_LEVEL_INFO:
                print(f"  Level {args.level}: {DEBUG_LEVEL_INFO[args.level]}")
            else:
                print(f"Level {args.level} is not valid. Debug mode levels are: {list(DEBUG_LEVEL_INFO.keys())}")

if __name__ == "__main__":
    main()



import os
import torch
from .utils import get_free_gpus, configs, modify_config_file

def launch(args):
    cpu = False
    if args.command == "debug":
        os.environ["ACCMT_DEBUG_MODE"] = str(args.level)

    if args.cpu:
        os.environ["ACCMT_CPU"] = "1"
        cpu = True

    gpus = args.gpus.lower()
    strat = args.strat
    file = args.file
    extra_args = " ".join(args.extra_args)

    if "." in strat:
        accelerate_config_file = strat
    else:
        accelerate_config_file = configs[strat]

    if not cpu:
        if not torch.cuda.is_available():
            raise ImportError("Could not run CLI: CUDA is not available on your PyTorch installation.")
        
        NUM_DEVICES = torch.cuda.device_count()

        gpu_indices = ""
        if gpus == "available":
            gpu_indices = ",".join(get_free_gpus(NUM_DEVICES))
        elif gpus == "all":
            gpu_indices = ",".join(str(i) for i in range(NUM_DEVICES))
        else:
            gpu_indices = gpus.removeprefix(",").removesuffix(",")

        if gpu_indices == "":
            raise RuntimeError("Could not get GPU indices. If you're using 'available' in 'gpus' "
                            "parameter, make sure there is at least one GPU free of memory.")

        if args.N != "0":
            if ":" in args.N:
                _slice = slice(*map(lambda x: int(x.strip()) if x.strip() else None, args.N.split(':')))
                gpu_indices = ",".join([str(i) for i in range(NUM_DEVICES)][_slice])
            else:
                gpu_indices = ",".join(str(i) for i in range(int(args.N)))

        # TODO: For now, we need to find a way to collect processes that are running on certain GPUs to verify if they're free to use.
        #if not args.ignore_warnings:
        #    gpu_indices_list = [int(idx) for idx in gpu_indices.split(",")]
        #    device_indices_in_use = []
        #    for idx in gpu_indices_list:
        #        if cuda_device_in_use(idx):
        #            device_indices_in_use.append(idx)
        #
        #    if len(device_indices_in_use) > 0:
        #        raise RuntimeError(
        #            f"The following CUDA devices are in use: {device_indices_in_use}."
        #             "You can ignore this warning via '--ignore-warnings'."
        #        )
        
        num_processes = len(gpu_indices.split(","))
    else:
        if args.N == "0":
            raise RuntimeError("When running on CPU, '-N' must specify the number of processes to run.")
        
        num_processes = int(args.N)
    
    modify_config_file(accelerate_config_file, num_processes)
    
    optimization1 = f"OMP_NUM_THREADS={os.cpu_count() // num_processes}" if args.O1 else ""
    cuda_prefix = f"CUDA_VISIBLE_DEVICES={gpu_indices}" if not cpu else ""

    cmd = (f"{optimization1} {cuda_prefix} "
            f"accelerate launch --config_file={accelerate_config_file} "
            f"{file} {extra_args}")
    
    os.system(cmd)
